//-----------------------------------------------------------------------------
// Copyright (c) 2007-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Dhpoware
{
    /// <summary>
    /// Custom vertex structure used for normal mapping.
    /// </summary>
    public struct NormalMappedVertex : IVertexType
    {
        public static readonly VertexDeclaration VertexDeclaration = new VertexDeclaration
        (
            new VertexElement(0, VertexElementFormat.Vector3, VertexElementUsage.Position, 0),
            new VertexElement(sizeof(float) * 3, VertexElementFormat.Vector2, VertexElementUsage.TextureCoordinate, 0),
            new VertexElement(sizeof(float) * 5, VertexElementFormat.Vector3, VertexElementUsage.Normal, 0),
            new VertexElement(sizeof(float) * 8, VertexElementFormat.Vector4, VertexElementUsage.Tangent, 0)
        );

        public Vector3 Position;
        public Vector2 TexCoord;
        public Vector3 Normal;
        public Vector4 Tangent;

        public NormalMappedVertex(Vector3 position, Vector2 texCoord, Vector3 normal, Vector4 tangent)
        {
            Position = position;
            TexCoord = texCoord;
            Normal = normal;
            Tangent = tangent;
        }

        VertexDeclaration IVertexType.VertexDeclaration
        {
            get { return VertexDeclaration; }
        }

        /// <summary>
        /// Given the 3 vertices (position and texture coordinate) and the
        /// face normal of a triangle calculate and return the triangle's
        /// tangent vector. This method is designed to work with XNA's default
        /// right handed coordinate system and clockwise triangle winding order.
        /// Undefined behavior will result if any other coordinate system
        /// and/or winding order is used. The handedness of the local tangent
        /// space coordinate system is stored in the tangent's w component.
        /// </summary>
        /// <param name="pos1">Triangle vertex 1 position</param>
        /// <param name="pos2">Triangle vertex 2 position</param>
        /// <param name="pos3">Triangle vertex 3 position</param>
        /// <param name="texCoord1">Triangle vertex 1 texture coordinate</param>
        /// <param name="texCoord2">Triangle vertex 2 texture coordinate</param>
        /// <param name="texCoord3">Triangle vertex 3 texture coordinate</param>
        /// <param name="normal">Triangle face normal</param>
        /// <param name="tangent">Calculated tangent vector</param>
        public static void CalcTangent(ref Vector3 pos1,
                                       ref Vector3 pos2,
                                       ref Vector3 pos3,
                                       ref Vector2 texCoord1,
                                       ref Vector2 texCoord2,
                                       ref Vector2 texCoord3,
                                       ref Vector3 normal,                                       
                                       out Vector4 tangent)
        {
            // Create 2 vectors in object space.
            // edge1 is the vector from vertex positions pos1 to pos3.
            // edge2 is the vector from vertex positions pos1 to pos2.
            Vector3 edge1 = pos3 - pos1;
            Vector3 edge2 = pos2 - pos1;
            
            edge1.Normalize();
            edge2.Normalize();

            // Create 2 vectors in tangent (texture) space that point in the
            // same direction as edge1 and edge2 (in object space).
            // texEdge1 is the vector from texture coordinates texCoord1 to texCoord3.
            // texEdge2 is the vector from texture coordinates texCoord1 to texCoord2.
            Vector2 texEdge1 = texCoord3 - texCoord1;
            Vector2 texEdge2 = texCoord2 - texCoord1;

            texEdge1.Normalize();
            texEdge2.Normalize();

            // These 2 sets of vectors form the following system of equations:
            //
            //  edge1 = (texEdge1.x * tangent) + (texEdge1.y * bitangent)
            //  edge2 = (texEdge2.x * tangent) + (texEdge2.y * bitangent)
            //
            // Using matrix notation this system looks like:
            //
            //  [ edge1 ]     [ texEdge1.x  texEdge1.y ]  [ tangent   ]
            //  [       ]  =  [                        ]  [           ]
            //  [ edge2 ]     [ texEdge2.x  texEdge2.y ]  [ bitangent ]
            //
            // The solution is:
            //
            //  [ tangent   ]        1     [ texEdge2.y  -texEdge1.y ]  [ edge1 ]
            //  [           ]  =  -------  [                         ]  [       ]
            //  [ bitangent ]      det A   [-texEdge2.x   texEdge1.x ]  [ edge2 ]
            //
            //  where:
            //        [ texEdge1.x  texEdge1.y ]
            //    A = [                        ]
            //        [ texEdge2.x  texEdge2.y ]
            //
            //    det A = (texEdge1.x * texEdge2.y) - (texEdge1.y * texEdge2.x)
            //
            // From this solution the tangent space basis vectors are:
            //
            //    tangent = (1 / det A) * ( texEdge2.y * edge1 - texEdge1.y * edge2)
            //  bitangent = (1 / det A) * (-texEdge2.x * edge1 + texEdge1.x * edge2)
            //     normal = cross(tangent, bitangent)

            Vector3 t;
            Vector3 b;
            float det = (texEdge1.X * texEdge2.Y) - (texEdge1.Y * texEdge2.X);

            if ((float)Math.Abs(det) < 1e-6f)    // almost equal to zero
            {
                t = Vector3.UnitX;
                b = Vector3.UnitY;
            }
            else
            {
                det = 1.0f / det;

                t.X = (texEdge2.Y * edge1.X - texEdge1.Y * edge2.X) * det;
                t.Y = (texEdge2.Y * edge1.Y - texEdge1.Y * edge2.Y) * det;
                t.Z = (texEdge2.Y * edge1.Z - texEdge1.Y * edge2.Z) * det;

                b.X = (-texEdge2.X * edge1.X + texEdge1.X * edge2.X) * det;
                b.Y = (-texEdge2.X * edge1.Y + texEdge1.X * edge2.Y) * det;
                b.Z = (-texEdge2.X * edge1.Z + texEdge1.X * edge2.Z) * det;

                t.Normalize();
                b.Normalize();
            }

            // Calculate the handedness of the local tangent space.
            // The bitangent vector is the cross product between the triangle face
            // normal vector and the calculated tangent vector. The resulting bitangent
            // vector should be the same as the bitangent vector calculated from the
            // set of linear equations above. If they point in different directions
            // then we need to invert the cross product calculated bitangent vector. We
            // store this scalar multiplier in the tangent vector's 'w' component so
            // that the correct bitangent vector can be generated in the normal mapping
            // shader's vertex shader.

            Vector3 bitangent = Vector3.Cross(normal, t);
            float handedness = (Vector3.Dot(bitangent, b) < 0.0f) ? -1.0f : 1.0f;

            tangent.X = t.X;
            tangent.Y = t.Y;
            tangent.Z = t.Z;
            tangent.W = handedness;
        }
    }


    /// <summary>
    /// The NormalMappedCube class is used to procedurally generate a cube.
    /// This class generates geometry using the NormalMappedVertex structure.
    /// </summary>
    public class NormalMappedCube
    {
        private float size;
        private NormalMappedVertex[] vertices;
        private VertexBuffer vertexBuffer;

        /// <summary>
        /// Constructs an instance of the NormalMappedCube class.
        /// </summary>
        /// <param name="graphicsDevice">The active GraphicsDevice object.</param>
        /// <param name="size">The size of the cube to generate.</param>
        public NormalMappedCube(GraphicsDevice graphicsDevice, float size)
        {
            Generate(size, graphicsDevice);
        }

        /// <summary>
        /// Draws the NormalMappedCube object using the shader contained within
        /// the passed in Effect object.
        /// </summary>
        /// <param name="effect">The Effect to render the cube with.</param>
        /// <param name="graphicsDevice">The active GraphicsDevice object.</param>
        public void Draw(Effect effect, GraphicsDevice graphicsDevice)
        {
            graphicsDevice.SetVertexBuffer(vertexBuffer);

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                graphicsDevice.DrawPrimitives(PrimitiveType.TriangleList, 0, 12);
            }
        }

        /// <summary>
        /// Returns the dimensions of the NormalMappedCube object.
        /// </summary>
        public float Size
        {
            get { return size; }
        }

        /// <summary>
        /// Procedurally generates a cube of the specified size.
        /// </summary>
        /// <param name="size">The size of the cube.</param>
        /// <param name="graphicsDevice">The active GraphicsDevice object.</param>
        private void Generate(float size, GraphicsDevice graphicsDevice)
        {
            /* 0--------1
             * |\      /|
             * | 4----5 |
             * | |    | |
             * | |    | |
             * | 7----6 |
             * |/      \|
             * 3--------2
             *
             * +z face: 0123 tri1: 012 tri2: 230
             * -z face: 5476 tri1: 547 tri2: 765
             *
             * +y face: 4510 tri1: 451 tri2: 104
             * -y face: 3267 tri1: 326 tri2: 673
             *
             * +x face: 1562 tri1: 156 tri2: 621
             * -x face: 4037 tri1: 403 tri2: 374
             */

            this.size = size;

            float halfSize = size * 0.5f;

            Vector3[] corners =
            {
                new Vector3(-halfSize,  halfSize,  halfSize),   // 0
                new Vector3( halfSize,  halfSize,  halfSize),   // 1
                new Vector3( halfSize, -halfSize,  halfSize),   // 2
                new Vector3(-halfSize, -halfSize,  halfSize),   // 3
                
                new Vector3(-halfSize,  halfSize, -halfSize),   // 4
                new Vector3( halfSize,  halfSize, -halfSize),   // 5
                new Vector3( halfSize, -halfSize, -halfSize),   // 6
                new Vector3(-halfSize, -halfSize, -halfSize),   // 7
            };

            Vector2[] texCoords =
            {
                new Vector2(0.0f, 0.0f),    // top left corner
                new Vector2(1.0f, 0.0f),    // top right corner
                new Vector2(1.0f, 1.0f),    // bottom right corner    
                new Vector2(0.0f, 1.0f),    // bottom left corner
            };

            vertices = new NormalMappedVertex[36];

            // +z face: 0123 tri1: 012 tri2: 230
            vertices[0] = new NormalMappedVertex(corners[0], texCoords[0], Vector3.Backward, Vector4.Zero);
            vertices[1] = new NormalMappedVertex(corners[1], texCoords[1], Vector3.Backward, Vector4.Zero);
            vertices[2] = new NormalMappedVertex(corners[2], texCoords[2], Vector3.Backward, Vector4.Zero);
            vertices[3] = new NormalMappedVertex(corners[2], texCoords[2], Vector3.Backward, Vector4.Zero);
            vertices[4] = new NormalMappedVertex(corners[3], texCoords[3], Vector3.Backward, Vector4.Zero);
            vertices[5] = new NormalMappedVertex(corners[0], texCoords[0], Vector3.Backward, Vector4.Zero);

            // -z face: 5476 tri1: 547 tri2: 765
            vertices[6] = new NormalMappedVertex(corners[5], texCoords[0], Vector3.Forward, Vector4.Zero);
            vertices[7] = new NormalMappedVertex(corners[4], texCoords[1], Vector3.Forward, Vector4.Zero);
            vertices[8] = new NormalMappedVertex(corners[7], texCoords[2], Vector3.Forward, Vector4.Zero);
            vertices[9] = new NormalMappedVertex(corners[7], texCoords[2], Vector3.Forward, Vector4.Zero);
            vertices[10] = new NormalMappedVertex(corners[6], texCoords[3], Vector3.Forward, Vector4.Zero);
            vertices[11] = new NormalMappedVertex(corners[5], texCoords[0], Vector3.Forward, Vector4.Zero);
            
            // +y face: 4510 tri1: 451 tri2: 104
            vertices[12] = new NormalMappedVertex(corners[4], texCoords[0], Vector3.Up, Vector4.Zero);
            vertices[13] = new NormalMappedVertex(corners[5], texCoords[1], Vector3.Up, Vector4.Zero);
            vertices[14] = new NormalMappedVertex(corners[1], texCoords[2], Vector3.Up, Vector4.Zero);
            vertices[15] = new NormalMappedVertex(corners[1], texCoords[2], Vector3.Up, Vector4.Zero);
            vertices[16] = new NormalMappedVertex(corners[0], texCoords[3], Vector3.Up, Vector4.Zero);
            vertices[17] = new NormalMappedVertex(corners[4], texCoords[0], Vector3.Up, Vector4.Zero);

            // -y face: 3267 tri1: 326 tri2: 673
            vertices[18] = new NormalMappedVertex(corners[3], texCoords[0], Vector3.Down, Vector4.Zero);
            vertices[19] = new NormalMappedVertex(corners[2], texCoords[1], Vector3.Down, Vector4.Zero);
            vertices[20] = new NormalMappedVertex(corners[6], texCoords[2], Vector3.Down, Vector4.Zero);
            vertices[21] = new NormalMappedVertex(corners[6], texCoords[2], Vector3.Down, Vector4.Zero);
            vertices[22] = new NormalMappedVertex(corners[7], texCoords[3], Vector3.Down, Vector4.Zero);
            vertices[23] = new NormalMappedVertex(corners[3], texCoords[0], Vector3.Down, Vector4.Zero);
            
            // +x face: 1562 tri1: 156 tri2: 621
            vertices[24] = new NormalMappedVertex(corners[1], texCoords[0], Vector3.Right, Vector4.Zero);
            vertices[25] = new NormalMappedVertex(corners[5], texCoords[1], Vector3.Right, Vector4.Zero);
            vertices[26] = new NormalMappedVertex(corners[6], texCoords[2], Vector3.Right, Vector4.Zero);
            vertices[27] = new NormalMappedVertex(corners[6], texCoords[2], Vector3.Right, Vector4.Zero);
            vertices[28] = new NormalMappedVertex(corners[2], texCoords[3], Vector3.Right, Vector4.Zero);
            vertices[29] = new NormalMappedVertex(corners[1], texCoords[0], Vector3.Right, Vector4.Zero);

            // -x face: 4037 tri1: 403 tri2: 374
            vertices[30] = new NormalMappedVertex(corners[4], texCoords[0], Vector3.Left, Vector4.Zero);
            vertices[31] = new NormalMappedVertex(corners[0], texCoords[1], Vector3.Left, Vector4.Zero);
            vertices[32] = new NormalMappedVertex(corners[3], texCoords[2], Vector3.Left, Vector4.Zero);
            vertices[33] = new NormalMappedVertex(corners[3], texCoords[2], Vector3.Left, Vector4.Zero);
            vertices[34] = new NormalMappedVertex(corners[7], texCoords[3], Vector3.Left, Vector4.Zero);
            vertices[35] = new NormalMappedVertex(corners[4], texCoords[0], Vector3.Left, Vector4.Zero);

            // Calculate the tangent space basis vectors for the cube faces.

            Vector4 tangent;

            for (int i = 0; i < 36; i += 3)
            {
                NormalMappedVertex.CalcTangent(
                    ref vertices[i].Position,
                    ref vertices[i + 1].Position,
                    ref vertices[i + 2].Position,
                    ref vertices[i].TexCoord,
                    ref vertices[i + 1].TexCoord,
                    ref vertices[i + 2].TexCoord,
                    ref vertices[i].Normal,
                    out tangent);

                vertices[i].Tangent = tangent;
                vertices[i + 1].Tangent = tangent;
                vertices[i + 2].Tangent = tangent;
            }

            // Create and fill the vertex buffer with the cube geometry.

            vertexBuffer = new VertexBuffer(graphicsDevice, typeof(NormalMappedVertex), vertices.Length, BufferUsage.WriteOnly);
            vertexBuffer.SetData(vertices);
        }
    }
}